@@ -159,7 +159,7 @@ struct input_block
       struct
         {
           FILE *file;           /* input file handle */
-          const char *name;     /* name of PREVIOUS input file */
+          char *name;           /* name of PREVIOUS input file */
           int lineno;           /* current line number for do */
           /* Yet another attack of "The curse of global variables" (sigh) */
           int out_lineno;        /* current output line number do */
@@ -181,7 +181,7 @@ typedef struct input_block input_block;
 
 
 /* Current input file name.  */
-const char *current_file;
+char *current_file;
 
 /* Current input line number.  */
 int current_line;
@@ -712,7 +712,8 @@ unget_input (int ch)
 
 /*------------------------------------------------------------------------.
 | skip_line () simply discards all immediately following characters, upto |
-| the first newline.                                                      |
+| the first newline.  All whitespaces following this newline are also     |
+| discarded.                                                              |
 `------------------------------------------------------------------------*/
 
 void
@@ -722,6 +723,20 @@ skip_line (void)
 
   while ((ch = next_char ()) != CHAR_EOF && ch != '\n')
     ;
+  while ((ch = next_char ()) != CHAR_EOF && (ch == '\t' || ch == ' '))
+    ;
+  unget_input (ch);
+}
+
+/*------------------------------------------------------------------------.
+| skip_buffer () discards current buffer.  This function is called by     |
+| <return> to exit current function.                                      |
+`------------------------------------------------------------------------*/
+
+void
+skip_buffer (void)
+{
+  pop_input ();
 }
 
 
@@ -847,6 +862,21 @@ input_init (void)
   
 }
 
+/*----------------------.
+| Deallocate obstacks.  |
+`----------------------*/
+
+void
+input_deallocate (void)
+{
+  xfree (array_current_line);
+  xfree (eolcomm.string);
+
+  obstack_free (&token_stack, 0);
+  obstack_free (&input_stack, 0);
+  obstack_free (&wrapup_stack, 0);
+}
+
 
 /*-------------------------------------------.
 | Functions to manipulate the syntax table.  |
