@@ -34,9 +34,8 @@
 #include "mp4h.h"
 
 static void expand_macro __P ((symbol *, read_type));
-static void expand_unknown_macro __P ((char *, read_type));
 static void expand_token __P ((struct obstack *, read_type, token_type, token_data *));
-static void collect_body __P ((symbol *, struct obstack *));
+static void collect_body __P ((symbol *, struct obstack *, struct obstack *));
 
 /* Current recursion level in expand_macro ().  */
 int expansion_level = 0;
@@ -80,16 +79,8 @@ expand_token (struct obstack *obs, read_type expansion, token_type t,
     {                           /* TOKSW */
     case TOKEN_EOF:
     case TOKEN_MACDEF:
-      break;
-
     case TOKEN_BGROUP:
-      if (expansion == READ_ATTR_BODY)
-        obstack_1grow (obs, CHAR_BGROUP);
-      break;
-
     case TOKEN_EGROUP:
-      if (expansion == READ_ATTR_BODY)
-        obstack_1grow (obs, CHAR_EGROUP);
       break;
 
     case TOKEN_QUOTED:
@@ -113,13 +104,11 @@ expand_token (struct obstack *obs, read_type expansion, token_type t,
 
       if (! IS_ALPHA (*text))
         {
-          expand_unknown_macro (text, expansion);
+          shipout_text (obs, TOKEN_DATA_TEXT (td), strlen (TOKEN_DATA_TEXT (td)));
           break;
         }
       sym = lookup_symbol (text, SYMBOL_LOOKUP);
-      if (sym == NULL)
-        expand_unknown_macro (text, expansion);
-      else if (SYMBOL_TYPE (sym) == TOKEN_VOID)
+      if (sym == NULL || SYMBOL_TYPE (sym) == TOKEN_VOID)
         shipout_text (obs, TOKEN_DATA_TEXT (td), strlen (TOKEN_DATA_TEXT (td)));
       else
         expand_macro (sym, expansion);
@@ -176,12 +165,11 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp)
 
               /* The argument MUST be finished, whether we want it or not.  */
               obstack_1grow (obs, '\0');
-              text = obstack_finish (obs);
 
               if (TOKEN_DATA_TYPE (argp) == TOKEN_VOID)
                 {
                   TOKEN_DATA_TYPE (argp) = TOKEN_TEXT;
-                  TOKEN_DATA_TEXT (argp) = text;
+                  TOKEN_DATA_TEXT (argp) = obstack_finish (obs);
                 }
               return (boolean) (IS_SPACE(*TOKEN_DATA_TEXT (&td)));
             }
@@ -288,7 +276,7 @@ collect_arguments (symbol *sym, read_type expansion, struct obstack *argptr,
       while (more_args);
       
       /*  If the last argument is empty, it is removed.  We need it to
-          remove wjite spaces before closing brackets.  */
+          remove white spaces before closing brackets.  */
       if (TOKEN_DATA_TYPE (tdp) == TOKEN_TEXT
           && strlen (TOKEN_DATA_TEXT (tdp)) == 0)
         argptr->next_free = last_addr;
@@ -299,16 +287,14 @@ collect_arguments (symbol *sym, read_type expansion, struct obstack *argptr,
 | Collect the body of a container tag. No expansion is performed.  |
 `-----------------------------------------------------------------*/
 static void
-collect_body (symbol *sym, struct obstack *argptr)
+collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
 {
   token_type t;
   token_data td;
   token_data *tdp;
-  struct obstack body;
   char *text;
   symbol *newsym;
 
-  obstack_init (&body);
   while (1)
     {
       t = next_token (&td, READ_BODY);
@@ -331,10 +317,10 @@ collect_body (symbol *sym, struct obstack *argptr)
         case TOKEN_SPACE:
         case TOKEN_STRING:
           if (expansion_level > 0 && t == TOKEN_QUOTED)
-            obstack_1grow (&body, CHAR_LQUOTE);
-          shipout_text (&body, text, strlen (text));
+            obstack_1grow (bodyptr, CHAR_LQUOTE);
+          shipout_text (bodyptr, text, strlen (text));
           if (expansion_level > 0 && t == TOKEN_QUOTED)
-            obstack_1grow (&body, CHAR_RQUOTE);
+            obstack_1grow (bodyptr, CHAR_RQUOTE);
           break;
 
         case TOKEN_WORD:
@@ -351,38 +337,40 @@ collect_body (symbol *sym, struct obstack *argptr)
                     {
                       t = next_token (&td, READ_BODY);
                     }
-                  while (! IS_CLOSE(*TOKEN_DATA_TEXT (&td)));
-                  obstack_1grow (&body, '\0');
+                  while (! IS_CLOSE(*TOKEN_DATA_TEXT (&td)))
+                    ;
+
+                  obstack_1grow (bodyptr, '\0');
 
                   /* Add body to argptr */
                   TOKEN_DATA_TYPE (&td) = TOKEN_TEXT;
-                  TOKEN_DATA_TEXT (&td) = obstack_finish (&body);
-                  tdp = (token_data *) obstack_copy (&body,
+                  TOKEN_DATA_TEXT (&td) = obstack_finish (bodyptr);
+                  tdp = (token_data *) obstack_copy (bodyptr,
                       (voidstar) &td, sizeof (td));
                   obstack_grow (argptr, (voidstar) &tdp, sizeof (tdp));
 
                   return;
-                  break;
                 }
               else
                 {
-                  obstack_grow (&body, TOKEN_DATA_TEXT (&td),
+                  obstack_grow (bodyptr, TOKEN_DATA_TEXT (&td),
                           strlen(TOKEN_DATA_TEXT (&td)) );
                   /*  gobble closing bracket */
                   do
                     {
                       t = next_token (&td, READ_BODY);
-                      obstack_grow (&body, TOKEN_DATA_TEXT (&td),
+                      obstack_grow (bodyptr, TOKEN_DATA_TEXT (&td),
                               strlen(TOKEN_DATA_TEXT (&td)) );
                     }
-                  while (! IS_CLOSE(*TOKEN_DATA_TEXT (&td)));
+                  while (! IS_CLOSE(*TOKEN_DATA_TEXT (&td)))
+                    ;
                 }
             }
           else
             {
               newsym = lookup_symbol (text, SYMBOL_LOOKUP);
               if (newsym == NULL || SYMBOL_TYPE (newsym) == TOKEN_VOID)
-                expand_unknown_macro (text, READ_ATTR_BODY);
+                shipout_text (bodyptr, TOKEN_DATA_TEXT (&td), strlen (TOKEN_DATA_TEXT (&td)));
               else
                 expand_macro (newsym, READ_ATTR_BODY);
             }
@@ -394,7 +382,6 @@ collect_body (symbol *sym, struct obstack *argptr)
           abort ();
         }
     }
-  obstack_free (&body, NULL);
 }
 
 
@@ -411,6 +398,9 @@ void
 call_macro (symbol *sym, struct obstack *obs, int argc, token_data **argv,
                  read_type expansion)
 {
+  if (SYMBOL_HOOK_BEGIN (sym))
+    obstack_grow (obs, SYMBOL_HOOK_BEGIN (sym),
+                        strlen (SYMBOL_HOOK_BEGIN (sym)));
   switch (SYMBOL_TYPE (sym))
     {
     case TOKEN_FUNC:
@@ -426,6 +416,9 @@ call_macro (symbol *sym, struct obstack *obs, int argc, token_data **argv,
         _("INTERNAL ERROR: Bad symbol type in call_macro ()")));
       abort ();
     }
+  if (SYMBOL_HOOK_END (sym))
+    obstack_grow (obs, SYMBOL_HOOK_END (sym),
+                        strlen (SYMBOL_HOOK_END (sym)));
 }
 
 /*-------------------------------------------------------------------------.
@@ -441,8 +434,7 @@ call_macro (symbol *sym, struct obstack *obs, int argc, token_data **argv,
 static void
 expand_macro (symbol *sym, read_type expansion)
 {
-  struct obstack arguments;
-  struct obstack argptr, save_argptr;
+  struct obstack arguments, argptr, save_argptr, body;
   token_data **argv;
   int argc, i;
   struct obstack *obs_expansion;
@@ -463,8 +455,9 @@ expand_macro (symbol *sym, read_type expansion)
 
   traced = (boolean) ((debug_level & DEBUG_TRACE_ALL) || SYMBOL_TRACED (sym));
 
-  obstack_init (&argptr);
   obstack_init (&arguments);
+  obstack_init (&argptr);
+  obstack_init (&body);
 
   if (traced && (debug_level & DEBUG_TRACE_CALL))
     trace_prepre (SYMBOL_NAME (sym), my_call_id);
@@ -483,7 +476,7 @@ expand_macro (symbol *sym, read_type expansion)
 
   if (SYMBOL_CONTAINER (sym))
     {
-      collect_body (sym, &save_argptr);
+      collect_body (sym, &save_argptr, &body);
       argv = (token_data **) obstack_finish (&save_argptr);
     }
 
@@ -507,7 +500,19 @@ expand_macro (symbol *sym, read_type expansion)
       for (i = 1; i < argc; i++)
         {
           obstack_1grow (obs_expansion, ' ');
-          shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
+          if (strlen (TOKEN_DATA_TEXT (argv[i])) == 0)
+            {
+              obstack_1grow (obs_expansion, CHAR_BGROUP);
+              obstack_1grow (obs_expansion, CHAR_EGROUP);
+            }
+          else if (strchr (TOKEN_DATA_TEXT (argv[i]), ' '))
+            {
+              obstack_1grow (obs_expansion, CHAR_BGROUP);
+              shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
+              obstack_1grow (obs_expansion, CHAR_EGROUP);
+            }
+          else
+            shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
         }
       obstack_1grow (obs_expansion, '>');
       if (SYMBOL_CONTAINER (sym))
@@ -527,92 +532,8 @@ expand_macro (symbol *sym, read_type expansion)
 
   obstack_free (&arguments, NULL);
   obstack_free (&argptr, NULL);
-}
+  obstack_free (&body, NULL);
 
-static void
-expand_unknown_macro (char *name, read_type expansion)
-{
-  symbol unknown;
-  struct obstack arguments;
-  struct obstack argptr;
-  token_data **argv;
-  int argc, i;
-  struct obstack *obs_expansion;
-  char *symbol_name, *cp;
-  const char *expanded;
-  boolean traced;
-  int my_call_id;
-
-  /* The ``name'' pointer points to a region which is modified
-     so we have to copy it now.  */
-  symbol_name = xstrdup (name);
-  expansion_level++;
-  if (expansion_level > nesting_limit)
-    MP4HERROR ((EXIT_FAILURE, 0,
-      _("ERROR: Recursion limit of %d exceeded, use -L<N> to change it"),
-           nesting_limit));
-
-  macro_call_id++;
-  my_call_id = macro_call_id;
-
-  /*  define a temporary symbol */
-  SYMBOL_NAME (&unknown)        = symbol_name;
-  SYMBOL_TRACED (&unknown)      = FALSE;
-  SYMBOL_CONTAINER (&unknown)   = FALSE;
-  SYMBOL_EXPAND_ARGS (&unknown) = TRUE;
-
-  traced = (boolean) ((debug_level & DEBUG_TRACE_ALL) || SYMBOL_TRACED (&unknown));
-
-  obstack_init (&argptr);
-  obstack_init (&arguments);
-
-  if (traced && (debug_level & DEBUG_TRACE_CALL))
-    trace_prepre (SYMBOL_NAME (&unknown), my_call_id);
-
-  if (expansion == READ_ATTR_BODY)
-    expansion = READ_ATTR_BODY;
-  else if (expansion == READ_ATTR_VERB || expansion == READ_BODY)
-    expansion = READ_ATTR_VERB;
-  else
-    expansion = READ_ATTRIBUTE;
-
-  collect_arguments (&unknown, expansion, &argptr, &arguments);
-  argc = obstack_object_size (&argptr) / sizeof (token_data *);
-  argv = (token_data **) obstack_finish (&argptr);
-
-  if (traced)
-    trace_pre (SYMBOL_NAME (&unknown), my_call_id, argc, argv);
-
-  obs_expansion = push_string_init ();
-  obstack_1grow (obs_expansion, '<');
-  obstack_grow (obs_expansion, symbol_name, strlen (symbol_name));
-
-  /*  Replace `name=val' attributes by name="val" */
-  for (i = 1; i < argc; i++)
-    {
-      obstack_1grow (obs_expansion, ' ');
-      cp = strchr (TOKEN_DATA_TEXT (argv[i]), '=');
-      if (cp != NULL && *(cp+1) != '"' && *(cp+1) != CHAR_QUOTE)
-        {
-          *cp = '\0';
-          shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
-          obstack_1grow (obs_expansion, '=');
-          obstack_1grow (obs_expansion, CHAR_QUOTE);
-          shipout_string (obs_expansion, cp + 1, 0);
-          obstack_1grow (obs_expansion, CHAR_QUOTE);
-        }
-      else
-        shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
-    }
-  obstack_1grow (obs_expansion, '>');
-  expanded = push_string_finish (READ_BODY);
-
-  if (traced)
-    trace_post (SYMBOL_NAME (&unknown), my_call_id, argc, argv, expanded);
-
-  --expansion_level;
-
-  obstack_free (&arguments, NULL);
-  obstack_free (&argptr, NULL);
-  xfree (symbol_name);
+  if (expansion_level == 0)
+    clear_tag_attr ();
 }
