@@ -268,6 +268,7 @@ extern FILE *debug;
   while (0)
 
 void debug_init __P ((void));
+void debug_deallocate __P ((void));
 int debug_decode __P ((const char *));
 void debug_flush_files __P ((void));
 boolean debug_set_output __P ((const char *));
@@ -342,9 +343,11 @@ typedef enum token_data_type token_data_type;
 typedef int read_type;
 
 void input_init __P ((void));
+void input_deallocate __P ((void));
 int peek_input __P ((void));
 token_type next_token __P ((token_data *, read_type));
 void skip_line __P ((void));
+void skip_buffer __P ((void));
 void input_close __P ((void));
 
 /* push back input */
@@ -359,7 +362,7 @@ boolean pop_wrapup __P ((void));
 void read_file_verbatim __P ((struct obstack *));
 
 /* current input file, and line */
-extern const char *current_file;
+extern char *current_file;
 extern int current_line;
 extern int *array_current_line;
 #define CURRENT_FILE_LINE current_file, array_current_line[expansion_level]
@@ -441,6 +444,7 @@ extern int current_diversion;
 extern int output_current_line;
 
 void output_init __P ((void));
+void output_deallocate __P ((void));
 void shipout_text __P ((struct obstack *, char *, int));
 void shipout_string __P ((struct obstack *obs, const char *s, int len));
 void make_diversion __P ((int));
@@ -471,6 +475,8 @@ struct symbol
   boolean expand_args;
 
   char *name;
+  char *hook_begin;
+  char *hook_end;
   token_data data;
 };
 
@@ -480,6 +486,8 @@ struct symbol
 #define SYMBOL_CONTAINER(S)     ((S)->container)
 #define SYMBOL_EXPAND_ARGS(S)   ((S)->expand_args)
 #define SYMBOL_NAME(S)          ((S)->name)
+#define SYMBOL_HOOK_BEGIN(S)    ((S)->hook_begin)
+#define SYMBOL_HOOK_END(S)      ((S)->hook_end)
 #define SYMBOL_TYPE(S)          (TOKEN_DATA_TYPE (&(S)->data))
 #define SYMBOL_TEXT(S)          (TOKEN_DATA_TEXT (&(S)->data))
 #define SYMBOL_FUNC(S)          (TOKEN_DATA_FUNC (&(S)->data))
@@ -493,6 +501,7 @@ typedef void hack_symbol ();
 extern symbol **symtab;
 
 void symtab_init __P ((void));
+void symtab_deallocate __P ((void));
 symbol *lookup_symbol __P ((const char *, symbol_lookup));
 symbol *lookup_variable __P ((const char *, symbol_lookup));
 void hack_all_symbols __P ((hack_symbol *, const char *));
@@ -517,17 +526,12 @@ struct builtin
   builtin_func *func;
 };
 
-struct predefined
-{
-  const char *unix_name;
-  const char *gnu_name;
-  const char *func;
-};
-
 typedef struct builtin builtin;
-typedef struct predefined predefined;
 
+void initialize_builtin __P ((symbol *));
 void builtin_init __P ((void));
+void builtin_deallocate __P ((void));
+void clear_tag_attr __P ((void));
 void define_builtin __P ((const char *, const builtin *, symbol_lookup,
                           boolean));
 void set_regexp_extended __P ((boolean));
@@ -567,8 +571,6 @@ struct search_path_info
   int max_length;               /* length of longest directory name */
 };
 
-struct search_path_info *search_path_info_new __P((void));
-void search_path_env_init __P ((struct search_path_info *, char *, boolean));
 void search_path_add __P ((struct search_path_info *, const char *));
 
 
